@@ -5,29 +5,22 @@ from collections import deque
 from enum import IntEnum
 from enum import IntFlag
 from enum import auto
-from functools import partial
-from functools import reduce
 from inspect import isawaitable
-from inspect import iscoroutinefunction
 from typing import TYPE_CHECKING
 from typing import Callable
 from typing import Dict
-from typing import Generator
-from typing import Iterable
 from typing import List
-from typing import Set
-from typing import Type
 
 from .exceptions import AttrNotFound
-from .exceptions import InvalidDefinition
 from .i18n import _
-from .spec_parser import custom_and
-from .spec_parser import operator_mapping
-from .spec_parser import parse_boolean_expr
 from .utils import ensure_iterable
 
 if TYPE_CHECKING:
-    from statemachine.dispatcher import Listeners
+    from typing import Set
+
+
+def allways_true(*args, **kwargs):
+    return True
 
 
 class CallbackPriority(IntEnum):
@@ -61,21 +54,6 @@ class CallbackGroup(IntEnum):
         return f"{self.name}@{id(specs)}"
 
 
-def allways_true(*args, **kwargs):
-    return True
-
-
-def take_callback(name: str, resolver: "Listeners", not_found_handler: Callable) -> Callable:
-    callbacks = list(resolver.search_name(name))
-    if len(callbacks) == 0:
-        not_found_handler(name)
-        return allways_true
-    elif len(callbacks) == 1:
-        return callbacks[0]
-    else:
-        return reduce(custom_and, callbacks)
-
-
 class CallbackSpec:
     """Specs about callbacks.
 
@@ -85,11 +63,15 @@ class CallbackSpec:
     before any real call is performed.
     """
 
+    names_not_found: "Set[str] | None" = None
+    """List of names that were not found on the model or statemachine"""
+
     def __init__(
         self,
         func,
         group: CallbackGroup,
         is_convention=False,
+        is_event: bool = False,
         cond=None,
         priority: CallbackPriority = CallbackPriority.NAMING,
         expected_value=None,
@@ -97,6 +79,7 @@ class CallbackSpec:
         self.func = func
         self.group = group
         self.is_convention = is_convention
+        self.is_event = is_event
         self.cond = cond
         self.expected_value = expected_value
         self.priority = priority
@@ -107,11 +90,22 @@ class CallbackSpec:
         elif callable(func):
             self.reference = SpecReference.CALLABLE
             self.is_bounded = hasattr(func, "__self__")
-            self.attr_name = func.__name__
+            self.attr_name = (
+                func.__name__ if not self.is_event or self.is_bounded else f"_{func.__name__}_"
+            )
+            if not self.is_bounded:
+                func.attr_name = self.attr_name
+                func.is_event = is_event
         else:
             self.reference = SpecReference.NAME
             self.attr_name = func
 
+        self.may_contain_boolean_expression = (
+            not self.is_convention
+            and self.group == CallbackGroup.COND
+            and self.reference == SpecReference.NAME
+        )
+
     def __repr__(self):
         return f"{type(self).__name__}({self.func!r}, is_convention={self.is_convention!r})"
 
@@ -127,68 +121,19 @@ class CallbackSpec:
     def __hash__(self):
         return id(self)
 
-    def _update_func(self, func: Callable, attr_name: str):
-        self.func = func
-        self.reference = SpecReference.CALLABLE
-        self.attr_name = attr_name
-
-    def _wrap(self, callback):
-        condition = self.cond if self.cond is not None else allways_true
-        return CallbackWrapper(
-            callback=callback,
-            condition=condition,
-            meta=self,
-            unique_key=callback.unique_key,
-        )
-
-    def build(self, resolver: "Listeners") -> Generator["CallbackWrapper", None, None]:
-        """
-        Resolves the `func` into a usable callable.
-
-        Args:
-            resolver (callable): A method responsible to build and return a valid callable that
-                can receive arbitrary parameters like `*args, **kwargs`.
-        """
-        if (
-            not self.is_convention
-            and self.group == CallbackGroup.COND
-            and self.reference == SpecReference.NAME
-        ):
-            names_not_found: Set[str] = set()
-            take_callback_partial = partial(
-                take_callback, resolver=resolver, not_found_handler=names_not_found.add
-            )
-            try:
-                expression = parse_boolean_expr(self.func, take_callback_partial, operator_mapping)
-            except SyntaxError as err:
-                raise InvalidDefinition(
-                    _("Failed to parse boolean expression '{}'").format(self.func)
-                ) from err
-            if not expression or names_not_found:
-                self.names_not_found = names_not_found
-                return
-            yield self._wrap(expression)
-            return
-
-        for callback in resolver.search(self):
-            yield self._wrap(callback)
-
 
 class SpecListGrouper:
-    def __init__(
-        self, list: "CallbackSpecList", group: CallbackGroup, factory=CallbackSpec
-    ) -> None:
+    def __init__(self, list: "CallbackSpecList", group: CallbackGroup) -> None:
         self.list = list
         self.group = group
-        self.factory = factory
         self.key = group.build_key(list)
 
     def add(self, callbacks, **kwargs):
-        self.list.add(callbacks, group=self.group, factory=self.factory, **kwargs)
+        self.list.add(callbacks, group=self.group, **kwargs)
         return self
 
     def __call__(self, callback):
-        return self.list._add_unbounded_callback(callback, group=self.group, factory=self.factory)
+        return self.list._add_unbounded_callback(callback, group=self.group)
 
     def _add_unbounded_callback(self, func, is_event=False, transitions=None, **kwargs):
         self.list._add_unbounded_callback(
@@ -196,7 +141,6 @@ class SpecListGrouper:
             is_event=is_event,
             transitions=transitions,
             group=self.group,
-            factory=self.factory,
             **kwargs,
         )
 
@@ -210,12 +154,13 @@ class CallbackSpecList:
     def __init__(self, factory=CallbackSpec):
         self.items: List[CallbackSpec] = []
         self.conventional_specs = set()
+        self._groupers: Dict[CallbackGroup, SpecListGrouper] = {}
         self.factory = factory
 
     def __repr__(self):
         return f"{type(self).__name__}({self.items!r}, factory={self.factory!r})"
 
-    def _add_unbounded_callback(self, func, is_event=False, transitions=None, **kwargs):
+    def _add_unbounded_callback(self, func, transitions=None, **kwargs):
         """This list was a target for adding a func using decorator
         `@<state|event>[.on|before|after|enter|exit]` syntax.
 
@@ -238,11 +183,7 @@ class CallbackSpecList:
                 event.
 
         """
-        spec = self._add(func, **kwargs)
-        if not getattr(func, "_specs_to_update", None):
-            func._specs_to_update = set()
-        if is_event:
-            func._specs_to_update.add(spec._update_func)
+        self._add(func, **kwargs)
         func._transitions = transitions
 
         return func
@@ -253,15 +194,16 @@ class CallbackSpecList:
     def clear(self):
         self.items = []
 
-    def grouper(
-        self, group: CallbackGroup, factory: Type[CallbackSpec] = CallbackSpec
-    ) -> SpecListGrouper:
-        return SpecListGrouper(self, group, factory=factory)
+    def grouper(self, group: CallbackGroup) -> SpecListGrouper:
+        if group not in self._groupers:
+            self._groupers[group] = SpecListGrouper(self, group)
+        return self._groupers[group]
 
-    def _add(self, func, group: CallbackGroup, factory=None, **kwargs):
-        if factory is None:
-            factory = self.factory
-        spec = factory(func, group, **kwargs)
+    def _add(self, func, group: CallbackGroup, **kwargs):
+        if isinstance(func, CallbackSpec):
+            spec = func
+        else:
+            spec = self.factory(func, group, **kwargs)
 
         if spec in self.items:
             return
@@ -291,7 +233,7 @@ class CallbackWrapper:
         unique_key: str,
     ) -> None:
         self._callback = callback
-        self._iscoro = iscoroutinefunction(callback)
+        self._iscoro = getattr(callback, "is_coroutine", False)
         self.condition = condition
         self.meta = meta
         self.unique_key = unique_key
@@ -338,19 +280,21 @@ class CallbacksExecutor:
     def __str__(self):
         return ", ".join(str(c) for c in self)
 
-    def _add(self, spec: CallbackSpec, resolver: "Listeners"):
-        for callback in spec.build(resolver):
-            if callback.unique_key in self.items_already_seen:
-                continue
+    def add(self, key: str, spec: CallbackSpec, builder: Callable[[], Callable]):
+        if key in self.items_already_seen:
+            return
 
-            self.items_already_seen.add(callback.unique_key)
-            insort(self.items, callback)
+        self.items_already_seen.add(key)
 
-    def add(self, items: Iterable[CallbackSpec], resolver: "Listeners"):
-        """Validate configurations"""
-        for item in items:
-            self._add(item, resolver)
-        return self
+        condition = spec.cond if spec.cond is not None else allways_true
+        wrapper = CallbackWrapper(
+            callback=builder(),
+            condition=condition,
+            meta=spec,
+            unique_key=key,
+        )
+
+        insort(self.items, wrapper)
 
     async def async_call(self, *args, **kwargs):
         return await asyncio.gather(
@@ -387,9 +331,6 @@ class CallbacksRegistry:
         self._registry: Dict[str, CallbacksExecutor] = defaultdict(CallbacksExecutor)
         self.has_async_callbacks: bool = False
 
-    def clear(self):
-        self._registry.clear()
-
     def __getitem__(self, key: str) -> CallbacksExecutor:
         return self._registry[key]
 
@@ -402,7 +343,8 @@ class CallbacksRegistry:
                 callback for callback in self[meta.group.build_key(specs)] if callback.meta == meta
             ):
                 continue
-            if hasattr(meta, "names_not_found"):
+
+            if meta.names_not_found:
                 raise AttrNotFound(
                     _("Did not found name '{}' from model or statemachine").format(
                         ", ".join(meta.names_not_found)
@@ -416,3 +358,24 @@ class CallbacksRegistry:
         self.has_async_callbacks = any(
             callback._iscoro for executor in self._registry.values() for callback in executor
         )
+
+    def call(self, key: str, *args, **kwargs):
+        if key not in self._registry:
+            return []
+        return self._registry[key].call(*args, **kwargs)
+
+    def async_call(self, key: str, *args, **kwargs):
+        return self._registry[key].async_call(*args, **kwargs)
+
+    def all(self, key: str, *args, **kwargs):
+        if key not in self._registry:
+            return True
+        return self._registry[key].all(*args, **kwargs)
+
+    def async_all(self, key: str, *args, **kwargs):
+        return self._registry[key].async_all(*args, **kwargs)
+
+    def str(self, key: str) -> str:
+        if key not in self._registry:
+            return ""
+        return str(self._registry[key])
